/*
*	based on /sys/src/cmd/strings.c
*/
#include	<u.h>
#include 	<libc.h>
#include 	<ctype.h>
#include	<bio.h>
#include	"misc.h"
#define DBG if(0)

Biobuf	bout;

#define	MINSPAN		2		/* Min characters in string (default) */
#define BUFSIZE		70

void dodir(int fd, char *path);
void dofile(int fd);
int isokrune(Rune);

static int minlen = MINSPAN;
static dflag;
static hflag;
static iflag;
static aflag;


char *spec = " !\"#$%&'()*+,-./0123456789:;<=>?@[\\]^_`{|}~";
/*
|20 sp |21  ! |22  " |23  # |24  $ |25  % |26  & |27  ' |
|28  ( |29  ) |2a  * |2b  + |2c  , |2d  - |2e  . |2f  / |
|30  0 |31  1 |32  2 |33  3 |34  4 |35  5 |36  6 |37  7 |
|38  8 |39  9 |3a  : |3b  ; |3c  < |3d  = |3e  > |3f  ? |
|40  @ 
|5b  [ |5c  \ |5d  ] |5e  ^ |5f  _ |
|60  ` 
|7b  { |7c  | |7d  } |7e  ~
*/



static void
usage(void)
{
	fprint(2, 
	"usage: %s [-l minlen] [-d] [-i] [-h] [-a] [path ...]\n"
	"  extwords extract words in files.\n"
	"  -d: the paths are directories\n"
	"  -l: minimum length of the words\n"
	"  -i: contents to lower case\n"
	"  -h: header name is last element of path\n"
	"  -hh: no header name\n"
	"  -a: add header info to contents (for indexing)\n"
	, argv0);
	exits("usage");
}


/*	we would like include file name to index. because file
*	name is basic information.
*	we want to avoid register symbols to the index.
*	but what to do? consider examples below:
*	.foo  _foo ._foo.tar	-> foo
*	1234.gz					-> discard
*	dots := dot + ...
*	name := {dots + nodotname} + ...
*	nodotname := {spec + word} + ...
*	we use only first word	*/

void
extfn(char *fn)
/*	fn is scratched */
{
	char *s, *t;
	s = skip(fn,".");
	t = skipto(s, "."); /* symbols include num, "." */
	*t = 0;
	s = skip(s, spec);
	t = skipto(s, spec);
	*t = 0;
	memcpy(fn, s, strlen(s) + 1);
}

char *
basename(char *path)
{
	char *p;
	p = strrchr(path,'/');
	if(p)
		p++;
	else
		p = path;
	return p;
}

void
main(int argc, char **argv)
{
	int i,n,fd;
	Dir *dir;
	char *p0,*p;

	ARGBEGIN{
	case 'd':
		dflag++;
		break;
	case 'l':
		minlen = atoi(EARGF(usage()));
		break;
	case 'h':
		hflag++;
		break;
	case 'a':
		aflag++;
		break;
	case 'i':
		iflag++;
		break;
	default:
		usage();
		break;
	}ARGEND

	Binit(&bout, 1, OWRITE);

	if(*argv == nil) {
		if(dflag){
			fd = open(".",OREAD);
			dodir(fd,".");
		}
		else
			dofile(0);
		exits(0);
	}



	if(dflag){
		for(i = 0; i < argc; i++) {
			/*	those argments may have trailing "/"
			*	if they have trailing "/", remove first	*/
			p = argv[i];
			n = strlen(p);
			if(p[n - 1] == '/')
				p[n - 1] = 0;
			fd = open(argv[i], OREAD);
			if(fd < 0)
				sysfatal("%s not open: %r",argv[i]);
			p0 = strdup(argv[i]);

			/*
			*	header output:
			*	case hflag
			*	0:  /dir_of_foo/foo:
			*	1:	foo:
			*	2:		# none
			*/

			if(!hflag)
				Bprint(&bout, "%s/:\n", p0);
			p = basename(p0);
			if(hflag == 1)
				Bprint(&bout, "%s/:\n", p);
			if(aflag){
				/*
				*	regulation to file name must be here
				*	it's late for mkmindex to do this jos
				*/
				if(iflag)
					strtolower(p);
				extfn(p);
				/* we may discard the name less than minlen */
				if(strlen(p) >= minlen) 
					Bprint(&bout, "%s\n", p);
			}
			free(p0);
			dir = dirfstat(fd);
			if(dir->mode & DMDIR) /* directory */
				dodir(fd, argv[i]);
			close(fd);
		}
	}
	else{
		for(i = 0; i < argc; i++) {
			fd = open(argv[i], OREAD);
			if(fd < 0)
				sysfatal("%s not open: %r",argv[i]);
			p0 = strdup(argv[i]);
			if(!hflag)
				Bprint(&bout, "%s:\n", p0);
			p = basename(p0);
			if(hflag == 1)
				Bprint(&bout, "%s:\n", p);
			if(aflag){	
				if(iflag)
					strtolower(p);
				extfn(p);
				/*	how should I do if the file name is 1.c	?
				*	the extfn(p) will return "" for the name.
				* we may discard the name less than minlen */
				if(strlen(p) >= minlen) 
					Bprint(&bout, "%s\n", p);
			}
			free(p0);
			dir = dirfstat(fd);
			if((dir->mode & DMDIR) == 0) /* file */
				dofile(fd);
			free(dir);
			close(fd);
		}
	}
	exits(0);
}

#define RUNEERROR 0xFFFD

int
compar(Dir *a, Dir *b)
{
	return strcmp(a->name, b->name);
}


void
dodir(int fd, char *path)
{
	Dir *dir;
	char *s, *p;
	int i,n;
	int fd1;

	n = dirreadall(fd,&dir);
	if(n == -1)
		sysfatal("dirreadall: %r");
	if(n == 0){
		free(dir);
		return;
	}

	/* sort by name */
	qsort(dir, n, sizeof(Dir), (int (*)(const void*, const void*))compar);

	for(i = 0; i < n; i++)
		if((dir[i].mode & DMDIR) == 0){
			s = smprint("%s/%s",path,dir[i].name);
			fd1 = open(s, OREAD);
			if(fd1 < 0)
				sysfatal("%s not open: %r", s);
			if(!hflag)
				Bprint(&bout, "%s:\n", s);
			p = dir[i].name;
			if(hflag == 1)
				Bprint(&bout, "%s:\n", p);
			if(aflag){	
				if(iflag)
					strtolower(p);
				extfn(p);
				/* we may discard the name less than minlen */
				if(strlen(p) >= minlen) 
					Bprint(&bout, "%s\n", p);
			}
			free(s);
			dofile(fd1);
			close(fd1);
		}
	free(dir);
}

void
dofile(int fd)
{
	Biobuf bin;
	int cnt = 0;
	long c;
	Rune buf[BUFSIZE];

	Binit(&bin, fd, OREAD);

	while((c = Bgetrune(&bin)) >= 0) {
		if(c == RUNEERROR)
			break;
		if(c != '\n' && c != '\r' && c != '\t' && iscntrl(c))
			break;
		if(isokrune(c)){
			if(iflag)
				c = tolower(c);
			buf[cnt++] = c;
			if(cnt == BUFSIZE-1) {
				buf[cnt] = 0;
				Bprint(&bout, "%S...\n", buf);
				cnt = 0;
			}
		}
		else{ /* space, digit, symbol, etc */
			if(cnt >= minlen) {
				buf[cnt] = 0;
				Bprint(&bout, "%S\n", buf);
			}
			cnt = 0;
		}
	}
	if(cnt >= minlen) {
		buf[cnt] = 0;
		Bprint(&bout, "%S\n", buf);
	}
	Bflush(&bout);
	Bterm(&bin);
}

int
isokrune(Rune r)
{
	if (r == Runeerror)
		return 0;
	if ((!(r & 0xff00) && isalpha(r)) || r > 0xA0)
		return 1;
	return 0;
}
