/*
*	grep -f
*	is buggy, so I wrote ffm
*	ffm: file to field maching
*	ffm is useful, because grep does not have concept for field
*	thus, it is hard or impossible for grep to handle some types of problems.
*	-Kenar-
*/

#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include "trans.h"
#include "misc.h"

#define DBG if(0)
#define	DBG0	if(0)
#define MAXFIELD 10
#define MAXBUF 1024

Biobuf bout;
char *sep;
int fnum;
int vflag;

typedef struct Hkeys Hkeys;
struct Hkeys{
	char *key;
	void *value;
	Hkeys *next;
} **hkeys;

/*	nhkey should be prime number
*	use command primes to find primes
*	primes 250 270
*	251
*	257
*	263
*	269		*/
int nhkeys =251;
int keylen;

/*	a simple linear hash */
int
mhash(char *s, int n)
{
	int h = 0;
	while(n-- > 0) h = (h << 1)%nhkeys + *s++;
	h %= nhkeys;
	return h<0?h+nhkeys:h;
}

int
hash(char *s)
{
	return mhash(s, strlen(s));
}


void
Hput(char *key, void *value)
{
	int h = hash(key);
	Hkeys *p = new(Hkeys);
	p->key = key;
	p->value = value;
	p->next = hkeys[h];
	hkeys[h] = p;
}

void *
Hget(char *key)
{
	Hkeys *p;
	for(p = hkeys[hash(key)];p;p = p->next)
		if(strcmp(p->key, key)==0)
			return p->value;
	return nil;
}

/*
*	patread: read translation data from pattern file	*/
void
patread(char *path)
{
	int n;
	char *p;
	List *q;
	List *lroot;

	n = filetolist(&lroot, path);
	nhkeys = upprime(n);
	DBG fprint(2,"eys=%d\n",nhkeys);
	hkeys = (Hkeys**) calloc(nhkeys, sizeof(Hkeys*));

	n = 0;
	for(q = lroot; q; q = q->next){
		p = q->data;
		Hput(p,"");
		n++;
	}
	if(n == 0)
		sysfatal("# no rule");
	DBG0 fprint(2,"done\n");
}

void
doit(int fd)
{
	char *p, *s, *args[MAXFIELD];
	char buf[MAXBUF];
	Rdbuf *rb;
	int m,n;
	rb = RdbufNew(fd, 32*1024);
	while((p = rdline(rb))){ /* assign = */
		/*	strncpy() is very slow.
		*	that does needless job. use strecpy instead	*/
		strecpy(buf, buf + MAXBUF, p);

		/*	Why getfields() is slower than tokenize() ?
		*	this is because getfields() allows utf coded chars
		*	in the separater. it would be much simpler one
		*	if the separater is a single char and ascii	*/
		if(sep)
			m = split(buf, args, MAXFIELD, *sep);
		else
			m = tokenize(buf, args, MAXFIELD);
		if(m < fnum)
			continue;

		if(fnum == 0){
			for(n = 0; n < m; n++){
				s = Hget(args[n]);
				if((!vflag && s) || (vflag && !s))
					Bprint(&bout, "%s\n", p);
			}
		}
		else{
			s = Hget(args[fnum-1]);
			if((!vflag && s) || (vflag && !s))
				Bprint(&bout, "%s\n", p);
		}
	}
	Bflush(&bout);
}

#ifdef XXX
/*
*	this isn't used, because it is too slow.
*	but I keep it so that I don't write it again.
*/
void
doit1(int fd)
{
	Biobufhdr bin;
	char *p, *s, *args[MAXFIELD];
	char buf[MAXBUF];
	uchar biobuf[32*1024];
	int m,n;

	if(Binits(&bin, fd, OREAD, biobuf, sizeof(biobuf)) < 0)
		sysfatal("input not open: %r");
	while((p = Brdstr(&bin,'\n',1))){ /* assign = */
		strncpy(buf,p,MAXBUF);
		if(sep)
			m = getfields(buf, args, MAXFIELD, 0, sep);
		else
			m = tokenize(buf, args, MAXFIELD);
		if(m < fnum)
			continue;
		if(fnum == 0){
			for(n = 0; n < m; n++){
				s = Hget(args[n]);
				if((!vflag && s) || (vflag && !s)) {
					Bwrite(&bout, p, strlen(p));
					Bwrite(&bout,"\n",1);
				}
			}
		}
		else{
			s = Hget(args[fnum-1]);
			if((!vflag && s) || (vflag && !s)){
				Bwrite(&bout, p, strlen(p));
				Bwrite(&bout,"\n",1);
			}
		}
	}
	Bflush(&bout);
}
#endif

void
usage(void)
{
	fprint(2,
	"usage: ffm [-f num] [-F sep] [-v] pat_file [file ....]\n"
	"	file to field maching\n"
	"	show lines if a field matches to one of strings in pat_file\n"
	" -f num: field number that begins with 1,2,...\n"
	" -F sep: field separator\n"
	" -v :    show lines that do not match the pattern.\n"
	"if num not be given, then fmp shows lines if one of the fields matches\n"
	"to one of strings in pat_file\n"
	);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int fd;
	ARGBEGIN{
	case 'F': sep = EARGF(usage());
		break;
	case 'f': fnum = atoi(EARGF(usage()));
		break;
	case 'v': vflag++;
		break;
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);

	patread(*argv++);

	if(*argv == nil){
		doit(0);
		exits(nil);
	}

	while(*argv){
		fd = open(*argv, OREAD);
		if(fd < 0)
			sysfatal("%s not open: %r",*argv);
		doit(fd);
		argv++;
	}
	exits(nil);
}
