/*
*	based on /sys/src/cmd/strings.c
*/
#include	<u.h>
#include 	<libc.h>
#include 	<ctype.h>
#include	<bio.h>

Biobuf	bout;

#define BUFSIZE		70

void dodir(int fd, char *path);
int dofile(int fd);
int isokrune(Rune);

static dflag;
static hflag;
static eflag;

static void
usage(void)
{
	fprint(2, "usage: %s [-d] [-h] [path ...]\n"
		"istext reports only text files in the argument\n"
		"options:\n"
		"  -d: the arguments are directory\n"
		"  -h: report only names (not full path)\n"
		"  -e: read to eof to report\n"
		, argv0);
	exits("usage");
}

void
main(int argc, char **argv)
{
	int i,fd;
	Dir *dir;
	char *p,*p0;
	int r;

	ARGBEGIN{
	case 'd':
		dflag++;
		break;
	case 'e':
		eflag++;
		break;
	case 'h':
		hflag++;
		break;
	default:
		usage();
		break;
	}ARGEND

	Binit(&bout, 1, OWRITE);

	if(*argv == nil) {
		if(dflag)
			usage();
		r = dofile(0);
		if(!r)
			print(":\n");
		exits(0);
	}

	if(dflag){
		for(i = 0; i < argc; i++) {
			fd = open(argv[i], OREAD);
			if(fd < 0)
				sysfatal("%s not open: %r",argv[i]);
			p0 = strdup(argv[i]);
			dir = dirfstat(fd);
			if(dir->mode & DMDIR) /* directory */
				dodir(fd, argv[i]);
			free(dir);
			free(p0);
			close(fd);
		}
	}
	else{
		for(i = 0; i < argc; i++) {
			fd = open(argv[i], OREAD);
			if(fd < 0)
				sysfatal("%s not open: %r",argv[i]);
			p0 = p = strdup(argv[i]);
			if(hflag){
				if(p[strlen(p) - 1] == '/')
					p[strlen(p) - 1] = 0;
				p = strrchr(argv[i],'/');
				if(p)
					p++;
				else
					p = p0;
			}
			dir = dirfstat(fd);
			if((dir->mode & DMDIR) == 0){ /* file */
				r = dofile(fd);
				if(!r)
					print("%s\n", p);
			}
			free(dir);
			free(p0);
			close(fd);
		}
	}
	exits(0);
}

#define RUNEERROR 0xFFFD

int
compar(Dir *a, Dir *b)
{
	return strcmp(a->name, b->name);
}

void
dodir(int fd, char *path)
{
	Dir *dir;
	char *s;
	int i,n,r;
	int fd1;

	n = dirreadall(fd,&dir);
	if(n == -1)
		sysfatal("dirreadall %r");

	/* sort by name */
	qsort(dir, n, sizeof(Dir), (int (*)(const void*, const void*))compar);

	for(i = 0; i < n; i++)
		if((dir[i].mode & DMDIR) == 0){
			s = smprint("%s/%s",path,dir[i].name);
			s = cleanname(s);
			fd1 = open(s, OREAD);
			r = dofile(fd1);
			if(!r){
				if(hflag)
					print("%s\n", dir[i].name);
				else
					print("%s\n", s);
			}
			close(fd1);
			free(s);
		}
	free(dir);
}

#define MAX 1000

int
dofile(int fd)
{
	Biobuf bin;
	int cnt;
	long c;

	Binit(&bin, fd, OREAD);

	cnt = 0;
	while((c = Bgetrune(&bin)) >= 0) {
		if(!isokrune(c))
			goto Lerr;
		cnt++;
		if(!eflag && cnt > MAX) 
			break;
	}
	Bterm(&bin);
	return 0;
Lerr:
	Bterm(&bin);
	if(eflag)
		fprint(2,"%llx %lx\n", Boffset(&bin),c);
	return 1;
}

int
isokrune(Rune r)
{
	if (r == Runeerror)
		return 0;
	if(r == '\n' || r == '\t' || r == '\r')
		return 1;
	if(r > 0xA0)
		return 1;
	if (isprint(r))
		return 1;
	return 0;
}
