/*
*	extract words from files
*/

#include <u.h>
#include <libc.h>
#include <bio.h>
#include "misc.h"

#define DBG if(0)

int minlen = 2;
int mincnt = 100000;
char *prefix = "ind_";

int qflag;

void
usage(void)
{
	fprint(2,"usage: mkmind [-p prefix] [-l minlen] [-m mincnt] index\n"
	"  mkmind splits database index into smaller pieces\n"
	"  index: index file which is sorted by dictionary order\n"
	"  prefix: prefic to created file (defaul ind_)\n",
	"  minlen: the smallest word length in the index (default 2)\n"
	"  mincnt: the minimum number of lines in the piece (default 100000)\n"
	);
	exits("usage");
}

void
doit(int fd)
{
	Biobuf *bop;
	int n;
	int maxlen;/* max length of index to be allowed */
	char *p,*q,*indx;
	char p0[256];
	char fname[256];
	Rdbuf *rb;
	int lprefix;
	p0[0] = 0;
	
	memcpy(fname,prefix,sizeof(fname));
	lprefix = strlen(prefix);
	indx = fname + lprefix;
	maxlen = sizeof(fname) - lprefix - 1;
	if(maxlen < minlen)
		sysfatal("so large prefix or minlen");
	indx[0] = 0;

	/* 	then fname is "ind_" for example	*/
	rb = RdbufNew(fd, 32*1024);
	bop = nil;
	n = 0;
	while((p = rdline(rb))){
		q = strchr(p,'\t'); /* data is separated by a single TAB */
		if(!q)
			sysfatal("no tab in line %s",p);
		*q = 0;

		if(!bop || (strlen(p) == minlen && strcmp(p,p0) != 0 && n >= mincnt)){
			/*	NB: the condition strlen(p) == minlen is critical.
			*	supppose minlen == 2.
			*	if we have a file
			*		ind_ab
			*		ind_cde
			*	and our request is "cd*", then both files can have
			*	words that match "cd*".
			*/
			if(bop)
				Bterm(bop);
			strecpy(indx, indx + minlen + 1, p); /* need room for '\0' */
			if(!qflag)
				print("# %s\n", fname);
			bop = Bopen(fname,OWRITE);
			if(!bop)
				sysfatal("%s cannot create: %r",fname);
			strecpy(p0,p0 + sizeof(p0), p);
			n = 0;
		}
		DBG print("%s\n", p);
		*q = '\t';
		DBG print("%s\n", p);
		if(bop)
			Bprint(bop,"%s\n", p);
		n++;
	}
	if(bop)
		Bterm(bop);
}

void
main(int argc, char *argv[])
{
	int fd;
	
	ARGBEGIN{
	case 'm': mincnt=atoi(EARGF(usage()));
		break;
	case 'l': minlen=atoi(EARGF(usage()));
		break;
	case 'p': prefix = strdup(EARGF(usage()));
		break;
	case 'q': qflag++;
		break;
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	fd = open(*argv,OREAD);
	if(fd < 0)
		sysfatal("%s not open: %r",*argv);
	doit(fd);
	exits(nil);
	close(fd);
}
