/*
*	mtest is match test
*	mtest test files if given pattern exist in files
*	NB: -l option is slow. don't use.
*	    instead make chain of grep. then ten times faster.
*		look G for example.
*	NB: mtest is not needed even if without -l option.
*	    grep is much faster in case we have may files to test,
*	    and also for large files.
*/

#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include <regexp.h>
#include "misc.h"

#define DBG if(0)

/*	MAX must be less or equal sizeof(long) */
#define MAX 32

int iflag; // ignore case
int lflag; // line wise test
int pflag; // print flag

Reprog *rep[MAX];

void
usage(void)
{
	fprint(2,"usage: mtest [-i] [-l] [-p] pattern [file ...]\n"
	"	pattern: re&re&...&re\n"
	"	it is assumed that re does not contain '&'\n"
	"options:\n"
	"  -i: ignore case\n"
	"  -l: linewise test\n",
	"  -p: print matched line (use with \"-l\" option)\n",
	"without \"-p\" option, output is only file names\n"
	);
	exits("usage");
}


/*	mkey() returns matched key no.
*	the key number is 0 to nre -1  */
int
mkey(Reprog **re, int nre, char *s)
{
	int i;
	Resub match[2];
	for(i = 0; i < nre; i++){
		match[0].sp = s;
		match[0].ep = nil;
		if(regexec(re[i], s, match, 2))
			break;
	}
	if(i == nre)
		return -1;
	return i;
}

int
addkey(long *mmap, int nre, char *mstr)
/*	nre is atmost sizeof(long) */
{
	int i,k,m;
	m = mkey(rep, nre, mstr);
	if(m < 0)
		return 0;
	*mmap = (*mmap | (1 << m));
	k = 0;
	for(i = 0; i < nre; i++)
		if(*mmap & (1 << i))
			k++;
	return k;
}



int
doit(char *name, Reprog *re, long *mmap, int nk)
{
	int  fd = 0;
	int lnum = 0;
	Rdbuf *rb;
	Resub match[2];
	char *p, *p0, *p1;
	int m,k;
	int mc = 0;
	char ch;
	USED(re);
	
	if(name){
		fd = open(name,OREAD);
		if(fd < 0)
			sysfatal("%s not open",name);
	}
	*mmap = 0;
	rb = RdbufNew(fd, 32*1024);
	while((p0 = p = rdline(rb))){
		lnum++;
		if(lflag)
			*mmap = 0;
		if(iflag){
			p1 = p = strdup(p0);
			strtolower(p);
		}
		/* don't foreget the next two lines */
		match[0].sp = p;
		match[0].ep = nil;
		while((m = regexec(re,p,match,2)) > 0){
			USED(m);
			DBG print("%d %s\n",m,p);
			DBG print("%s\n%s\n", match[0].sp, match[0].ep);
			ch = *(match[0].ep);
			*(match[0].ep) = 0;
			DBG print("%s\n", match[0].sp);
			/*	we don't know which key is matched. */
			k = addkey(mmap,nk,match[0].sp);
			if(k == nk){ /* all keys are processed */
				if(pflag){
					*(match[0].ep) = ch;
					print("%s:%d:%s\n",name,lnum,p0);
					DBG print("# %o\n", *mmap);
					mc++;
					if(lflag)
						break;
				}
				if(iflag)
					free(p1);
				RdbufFree(rb);
				return 1;
			}
			*(match[0].ep) = ch;
			p = match[0].ep;
			match[0].sp = p;
			match[0].ep = nil;
		}
		if(iflag)
			free(p1);
	}
	RdbufFree(rb);
	if(fd)
		close(fd);
	return mc;
}

void
main(int argc, char *argv[])
{
	int i,n,r;
	Reprog *re;
	char *args[MAX];
	char *e,*s;
	char rere[256];
	long mmap;
	int nm=0;

	ARGBEGIN{
	case 'i': iflag++;
		break;
	case 'l': lflag++;
		break;
	case 'p': pflag++;
		break;
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	s = strdup(argv[0]);
	if(iflag)
		strtolower(s);
	n = getfields(s, args, MAX, 1, "&");

	/*
	*	let argv[0] be r0&r1&...&rn
	*	we want to have
	*	(r0)|(r1)|...|(rn)
	*	this will be rere
	*/

	s = rere;
	e = rere + sizeof(rere);
	for(i = 0; i < n; i++){
		DBG print("%d %s\n", i, args[i]);
		s = seprint(s,e,"(%s)|",args[i]);
		DBG print("%s\n", rere);
	}
	rere[strlen(rere) - 1] = 0;
	DBG print("%s\n", rere);


	re = regcomp(rere);
	if(!re)
		sysfatal("regcomp");
	for(i = 0; i < n; i++)
		rep[i] = regcomp(args[i]);

	argv++;
	if(*argv == nil){
		r = doit(nil, re, &mmap,n);
		if(r){
			print("matched\n");
			exits(nil);
		}
		print("no match\n");
		exits("no match");
	}

	while(*argv){
		r = doit(*argv, re, &mmap,n);
		if(r){
			if(!pflag)
				print("%s\n",*argv);
			nm++;
		}
		argv++;
	};
	if(nm)
		exits(nil);
	exits("no match");
}
