#include <u.h>
#include <libc.h>
#include <bio.h>

#include "trans.h"


void *
emalloc(long n)
{
	void *p = malloc(n);
	if(p==0)
		sysfatal("Can't malloc %d bytes", n);
	return p;
}

char *
estrdup(char *str)
{
	char *p;
	p = strdup(str);
	if(p == nil)
		sysfatal("no space for strdup");
	return p;
}


int
mpow(int n, int k, int mod)
{
	int m;
	if(k == 0)
		return 1;
	m = mpow(n,k/2,mod);
	if(k%2)
		return m*m*n %mod;
	else
		return m*m %mod;
}


int
upprime(int n)
/*	get smallest prime >= n
*	applicable for small number */
{
	int ng,i,k;
	if(n%2 == 0)
		n++;
	for(i = n; ; i += 2){
		/* Fermat test */
		if(mpow(2,i-1,i) != 1) /* not a prime */
			continue;
		/* exact test */
		ng = 0;
		for(k = 3; k*k <= i; k += 2)
			if(i%k == 0){
				ng = 1;
				break;
			}
		if(ng)
			continue;
		return i;
	}
}

List *
Lpush(List *root, void *data)
{
	List *p = new(List);
	p->data = data;
	p->next = root;
	return p;
}

List*
Lappend(List *p, void *d)
{
	List *q;
	q = (List*) malloc(sizeof(List));
	q->next = p->next;
	q->data = d;
	p->next = q;
	return q;
};

void
Lfree(List *p)
{
	List *q;
	q = p->next;
	free(p);
	while(q != 0){
		p = q->next;
		free(q);
		q = p;
	}
}

int
filetolist(List **rootp, char *path)
{
	int fd,n;
	Biobuf bin;
	List *q;
	char *p;
	fd = open(path, OREAD);
	if(fd < 0)
		sysfatal("%s not open %r", path);
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("error in Binit: %r");
	n = 0;
	p = Brdstr(&bin, '\n', 1);
	if(!p)
		sysfatal("empty rule");
	n++;
	*rootp = q = Lpush(nil,p);
	while((p = Brdstr(&bin, '\n', 1))){ /* assign = */
		q = Lappend(q, estrdup(p));
		n++;
	}
	close(fd);
	return n;
}

int
listtotable(Ttab *ttab, List *root)
/*	return: number of rules in the table */
{
	int i,m;
	int fmt = 0;
	char *p;
	List *q;
	char *args[3];
	Ttab *tp;
	p = root->data;
	if(*p == '#'){
		if(cistrstr(p,"unicode"))
			fmt = ufmt;
		else
		if(cistrstr(p,"hex"))
			fmt = hfmt;
	}

	for(i = 0, q = root; q; q = q->next){
		p = q->data;
		if(*p == '#')
			continue;
		m = tokenize(p,args, 3);
		if(m < 2)
			continue;
		tp = &ttab[i];
		tp->frstr = trfmt(args[0],fmt);
		tp->tostr = trfmt(args[1],fmt);
		tp->frlen = strlen(tp->frstr);
		if(tp->frlen == 0)
			sysfatal("# null string in the first field\n");
		i++; /* don't move i++ to for(;;) */
	}
	return i;
}

int _ungetcount;
uchar _ungetbuf[512];

/*
 * Bgetch: extended Bgetc()
 */
int
Bgetch(Biobuf *bin)
{
	int c;
	if(_ungetcount > 0)
		return _ungetbuf[--_ungetcount];
	c = Bgetc(bin);
	if(c < 1)
		return ERR;
	return (unsigned)c&0xff;
}

/*
 * Bungetch: unget n chars.
 */
void
Bungetch(Biobuf *bip, char *s, int n)
{
	USED(bip); // dummy
	while(n > 0)
		_ungetbuf[_ungetcount++] = s[--n];
}



char *
ewrite(char *wp, char *ep, char *s, int n)
/*
*	wp: working point
*	ep: end point
*	return: new working point	*/
{
	if(wp + n > ep)
		n = ep - wp;
	memcpy(wp,s,n);
	wp += n;
	return wp;
}



char *
trfmt(char *str, int fmt)
/*	tr translate various formatted rule data to utf-8 raw format.
*	str: input
*	fmt: format style
*	return: converted str.
*	NB: str will be scratched */
{
	char *s,*end;
	s = str;

	if(fmt == hfmt){
		int i,ch;
		for(i = 0;;){
			ch = strtol(s,&end,16);
			str[i++] = ch;
			if(*end == 0)
				break;
			s = end+1;
		}
		str[i] = 0;
		return str;
	}
	if(fmt == ufmt){
		int i,n;
		Rune rune;
		for(i = 0;;){
			rune = strtol(s,&end,16);
			/* typedef	ushort		Rune;	*/
			n = runetochar(&str[i],&rune);
			i += n;
			if(*end == 0)
				break;
			s = end+1;
		}
		str[i] = 0;
		return str;
	}
	return str;
}


/*	man strcat(2): 
*	The comparisons are made with unsigned bytes.	*/
int
ncompar(Ttab *x, Ttab *y)
/*	normal order comparison	*/
{
	return strcmp(x->frstr,y->frstr);
}

int
rcompar(Ttab *x, Ttab *y)
/*	reverse order comparison	*/
{
	return -strcmp(x->frstr,y->frstr);
}



/*
*	tabread: read translation data from rule file
*	lines beginning with "#" are comment
*	if the first line is comment and string "unicode" or "hex" is in it,
*	these are keywords for format of the file 	*/
int
readtable0(Trtab *trtab, char *path, int sortf)
/*	sortf: sort flag (-1,0,1)	*/
{
	int n,m;
	List *lroot;
	int max,min,i;
	Ttab *ttab;
	int (*compar)();

	n = filetolist(&lroot, path);
	ttab = calloc(n + 1, sizeof(Ttab));
	/* need room for nil */

	n = listtotable(ttab, lroot);
	/* n is num of element. don't change the meaning hereafter */
	if(n == 0)
		sysfatal("# no rule");
	if(sortf != 0){
		compar = (sortf>0?ncompar:rcompar);
		qsort(ttab, n, sizeof(Ttab), (int (*)(const void*, const void*))compar);
	}
	max = 0;
	min = 99999;
	for(i = 0; i < n; i++){
		m = ttab[i].frlen;
		if(max < m)
			max = m;
		if(min > m)
			min = m;
	}

	trtab->ttab = ttab;
	trtab->min = min;
	trtab->max = max;
	return n;
}

